/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef CallTree_H
#define CallTree_H

#include "Tree.h"
#include "AggregatedTreeItem.h"
#include "CubeTypes.h"

namespace cubegui
{
class TreeItem;
class CallTreeData;

class CallTree : public Tree
{
    Q_OBJECT
public:
    CallTree();
    ~CallTree();

    virtual QString
    getLabel();

    TreeItem*
    getTreeItem( uint32_t cnodeId );

    // ----  iteration handling
    /** define current tree item as root item of a loop */
    void
    setAsLoop( TreeItem* item );

    /** hide iterations, show aggregated subtree */
    void
    hideIterations();

    /** show all iterations */
    void
    showIterations();

    /** true, if iterations are aggregated */
    bool
    iterationsAreHidden();

    /** return the root item of the current loop */
    TreeItem*
    getLoopRootItem();

    /** return the root item of the aggregated loop */
    AggregatedTreeItem*
    getAggregatedRootItem();

    // ----  end iteration handling

    void
    setItemAsRoot( TreeItem* item );
    void
    pruneItem( TreeItem* item );
    void
    setItemAsLeaf( TreeItem* item );

protected:
    virtual TreeItem*
    createTree();

    virtual QString
    getItemName( cube::Vertex* vertex );

    /**
     * Set the basic inclusive and exclusive values in the full tree.
     * @param left  List of trees to the left of this tree
     * @param right List of trees to the right of this tree
     * @param itemsToCalculate, if list is not empty, only the given items are calculated, otherwise all visible items are calculated
     */
    virtual QVector<Task*>
    setBasicValues( const QList<Tree*>&     leftTrees,
                    const QList<Tree*>&     rightTrees,
                    const QList<TreeItem*>& itemsToCalculate = QList<TreeItem*>( )
                    );

    virtual void
    assignValues( TreeItem*                          item,
                  const std::vector< cube::Value* >& inclusive_values,
                  const std::vector< cube::Value* >& exclusive_values,
                  int                                depth = INT_MAX,
                  cube::IdIndexMap*                  indexMap = NULL );

private:
    /* calltree can aggregate loop iterations */
    TreeItem*           loopRoot;           // item which is handled as a root of a loop and whose children are iterations
    AggregatedTreeItem* aggregatedLoopRoot; // root of the aggregated items of the currenly defined loop
    bool                iterationsAreHidden_;

    void
    updateTreeItemList();

    // ---- iteration handling ----
    void
    unsetLoop();

    /** returns true, if the item is the root item of an aggregated loop (with hidden items) */
    void
    deleteMergedIterations( TreeItem* iter );
    void
    mergeIterations( TreeItem*               newParent,
                     const QList<TreeItem*>& iterations );
    AggregatedTreeItem*
    mergeIterations( TreeItem* loopRoot );

    // ---- end iteration handling ----

    friend class CallTreeData;
};

/** for use in QConcurrent */
class CallTreeData : public Task
{
public:
    CallTreeData( TreeItem* item_, const cube::list_of_metrics& m, const cube::list_of_sysresources& s, cube::Cnode* root, size_t depth_ ) :
        item( item_ ), metric_selection( m ), sysres_selection( s ), root_cnode( root ), depth( depth_ )
    {
    }

    virtual void
    calculate();

private:
    TreeItem*                        item;
    const cube::list_of_metrics      metric_selection;
    const cube::list_of_sysresources sysres_selection;
    cube::Cnode*                     root_cnode;
    size_t                           depth;
};
}
#endif // CallTree_H
