/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2019                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"

#include "FlatTree.h"
#include "CubeMetric.h"
#include "CubeRegion.h"
#include "TreeView.h"
#include "CubeProxy.h"

using namespace std;
using namespace cube;
using namespace cubegui;

FlatTree::FlatTree()
{
    treeType = CALLFLAT;
}

QString
FlatTree::getLabel()
{
    return "Flat view";
}

QString
FlatTree::getItemName( cube::Vertex* vertex )
{
    return QString( ( ( cube::Region* )vertex )->get_name().c_str() );
}

/**
 * @brief FlatTree::createTree
 * creates a flat tree that has only one level of items.
 */
TreeItem*
FlatTree::createTree()
{
    TreeItem* top = new TreeItem( this, QString(), REGIONITEM, 0 );

    createItems<cube::Region>( top, cube->getRegions(), REGIONITEM );

    const vector<cube::Cnode*>& cnodeVector = cube->getCnodes();
    QSet<cube::Vertex*>         callerItems;
    foreach( Cnode * cnode, cnodeVector )
    {
        if ( cnode )
        {
            cube::Vertex* caller = cnode->get_caller();
            callerItems.insert( caller );
        }
    }

    // create a Subroutine subitem, if the item has any callees
    foreach( TreeItem * item, top->getChildren() )
    {
        cube::Vertex* region = item->getCubeObject();
        if ( callerItems.contains( region ) )
        {
            TreeItem* child = new TreeItem( this, "Subroutines", REGIONITEM, 0 );
            item->addChild( child );
            treeItemHash.insertMulti( 0, child );
        }
    }

    return top;
}

void
FlatTree::assignValues( TreeItem*               item,
                        const vector< Value* >& inclusive_values,
                        const vector< Value* >& exclusive_values,
                        const vector< Value* >& difference_values )
{
    assert( item );
    assert( item->getCubeObject() );
    assert( item->getCubeObject()->get_id() < inclusive_values.size() );
    assert( item->getCubeObject()->get_id() < exclusive_values.size() );

    item->setValues( inclusive_values[ item->getCubeObject()->get_id() ],
                     exclusive_values[ item->getCubeObject()->get_id() ] );

    // Assign value to subroutine-child, if it exists
    foreach( TreeItem * item_subroutine, item->getChildren() )
    {
        item_subroutine->setValues( difference_values[ item->getCubeObject()->get_id() ],
                                    difference_values[ item->getCubeObject()->get_id() ]->copy() );
    }
}

QVector<Task*>
FlatTree::setBasicValues( const QList<Tree*>&     leftTrees,
                          const QList<Tree*>&     rightTrees,
                          const QList<TreeItem*>& itemsToCalculate = QList<TreeItem*>( ) )
{
    QVector<Task*> workerData;
    /* No update of tree is necessary if the flat tree is to the left of the metric tree */
    if ( ( leftTrees.size() == 0 ) || ( ( leftTrees.size() == 1 ) && ( leftTrees[ 0 ]->getTreeType() != METRICTREE ) ) )
    {
        return workerData;
    }

    if ( itemsToCalculate.size() == 0 ) // all visible items are already up to date
    {
        return workerData;
    }

    // Set up tree pointers
    Tree* metricTree        = 0;
    Tree* systemTree        = 0;
    bool  rightOfSystemTree = false;

    // Metric tree is definitely in the left trees, as the initial if-statement at the
    // beginning of the call already filtered out the other cases.
    foreach( Tree * item, leftTrees )
    {
        if ( item->getTreeType() == METRICTREE )
        {
            metricTree = item;
        }
        else if ( item->getTreeType() == SYSTEMTREE )
        {
            systemTree        = item;
            rightOfSystemTree = true;
        }
    }
    assert( metricTree );

    // if system tree was not in the left trees, search in the right trees
    if ( !systemTree )
    {
        foreach( Tree * item, rightTrees )
        {
            if ( item->getTreeType() == SYSTEMTREE )
            {
                systemTree = item;
            }
        }
    }
    assert( systemTree );

    // Set up selection for metric tree
    const QList<TreeItem*>& selected_metrics = metricTree->getSelectionList();
    list_of_metrics         metric_selection;

    foreach( TreeItem * item, selected_metrics )
    {
        metric_pair mp;
        mp.first  = static_cast<cube::Metric*> ( item->getCubeObject() );
        mp.second = ( item->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
        if ( item->getChildren().size() == 0 )
        {
            mp.second = CUBE_CALCULATE_INCLUSIVE;
        }
        metric_selection.push_back( mp );
    }

    // Set up selection for system tree
    list_of_sysresources sysres_selection;
    if ( rightOfSystemTree )  // call tree is on the right
    {
        QList< TreeItem* > selected_systemnodes = systemTree->getSelectionList();
        foreach( TreeItem * item, selected_systemnodes )
        {
            sysres_pair sp;
            sp.first  = static_cast<cube::Sysres*> ( item->getCubeObject() );
            sp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
            sysres_selection.push_back( sp );
        }
    }
    else
    {
        sysres_selection = CubeProxy::ALL_SYSTEMNODES;
    }

    workerData.append( new FlatTreeData( this,
                                         metric_selection,
                                         sysres_selection ) );

    return workerData;
}

/** used to calculate the data of root_cnode in a thread */
void
FlatTreeData::calculate()
{
    vector< Value* > inclusive_values;
    vector< Value* > exclusive_values;
    vector< Value* > difference_values;

    cube::CubeProxy* cube = tree->getCube();

    // get values for full tree
    cube->getFlatTreeValues( metric_selection,
                             sysres_selection,
                             inclusive_values,
                             exclusive_values,
                             difference_values );

    foreach( TreeItem * child, tree->getTop().getChildren() )
    {
        // Assign values to individual nodes
        tree->assignValues( child, inclusive_values, exclusive_values, difference_values );
    }
}
