/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"

#include "SystemTree.h"
#include "CallTree.h"
#include "TreeView.h"
#include "CubeMetric.h"
#include "CubeProxy.h"
#include <QDebug>

using namespace cube;
using namespace std;
using namespace cubegui;

SystemTree::SystemTree()
{
    treeType = SYSTEMTREE;
}

QString
SystemTree::getLabel()
{
    //std::string title = cube->get_systemtree_title(); todo proxy
    //return ( title.compare( "" ) == 0 ) ? tr( "System tree" ) : QString::fromStdString( title );
    return tr( "System tree" );
}

/**
 * creates a system tree
 * The first level of the system tree consists of items of type SYSTEMTREENODEITEM. The children
 * of this level can be also SYSTEMTREENODEITEM.
 * The last two levels of the tree are items of type LOCATIONGROUPITEM which have children of
 * type LOCATION.
 */
TreeItem*
SystemTree::createTree()
{
    TreeItem*                  top = new TreeItem( this, QString(), SYSTEMTREENODEITEM, 0 );
    std::vector<cube::Vertex*> vec;

    // insert root systen tree nodes and its children recursively
    createItems( top, cube->getRootSystemTreeNodes(), SYSTEMTREENODEITEM );

    // insert LOCATIONGROUP items but not its children of type LOCATION
    QList<TreeItem*> groups;
    createItems( top, cube->getLocationGroups(), LOCATIONGROUPITEM, &groups, false );

    // insert LOCATION items
    createItems( top, cube->getLocations(), LOCATIONITEM );

    // for single-threaded system trees : hide the threads
    bool singleThreaded = true;
    foreach( TreeItem * group, groups ) // for all items of type LOCATIONGROUPITEM
    {
        if ( group->getChildren().size() > 1 )
        {
            singleThreaded = false;
            break;
        }
    }
    if ( singleThreaded )
    {
        foreach( TreeItem * group, groups )
        {
            if ( group->getChildren().size() == 1 )
            {
                TreeItem* child = group->child( 0 );
                group->removeChild( child );
                treeItemHash.remove( child->getCubeObject() );
            }
        }
    }
    // end handling of single-threaded trees

    return top;
}

QString
SystemTree::getItemName( cube::Vertex* vertex )
{
    return QString( ( ( cube::Sysres* )vertex )->get_name().c_str() );
}

void
SystemTree::computeMinMax()
{
    double         tmp = 0.0;
    vector<double> min_vector;
    vector<double> max_vector;
    vector<bool>   depth_seen;
    foreach( TreeItem * item, getItems() )
    {
        size_t depth = item->getDepth();

        if ( min_vector.size() <= depth )
        {
            min_vector.resize( depth + 1 );
        }
        if ( max_vector.size() <= depth )
        {
            max_vector.resize( depth + 1 );
        }
        while ( depth_seen.size() <= depth )
        {
            depth_seen.push_back( false );
        }

        tmp = item->totalValue;

        if ( !depth_seen[ depth ] )
        {
            depth_seen[ depth ] = true;
            min_vector[ depth ] = tmp;
            max_vector[ depth ] = tmp;
        }
        else
        {
            min_vector[ depth ] = ( min_vector[ depth ] > tmp ) ? tmp : min_vector[ depth ];
            max_vector[ depth ] = ( max_vector[ depth ] < tmp ) ? tmp : max_vector[ depth ];
        }
    }

    foreach( TreeItem * item, getItems() )
    {
        int depth = item->getDepth();
        item->minValue = min_vector[ depth ];
        item->maxValue = max_vector[ depth ];
    }
}


void
SystemTree::assignValues( TreeItem*               item,
                          const vector< Value* >& inclusive_values,
                          const vector< Value* >& exclusive_values )
{
    assert( item );
    assert( item->getCubeObject() );

    SystemTreeNode* stn = static_cast<SystemTreeNode*>( item->getCubeObject() );
    uint32_t        id  = stn->get_sys_id();
    // value vector is empty, if selected left tree item is zero -> set to NULL
    item->setValues( ( ( id >= inclusive_values.size() ) ? NULL : inclusive_values[ id ] ),
                     ( ( id >= exclusive_values.size() ) ? NULL : exclusive_values[ id ] ) );

    foreach( TreeItem * child, item->getChildren() )
    {
        assignValues( child, inclusive_values, exclusive_values );
    }
}

/**
   returns true, if all items have to be recalculated and calculation isn't disabled
 */
bool
SystemTree::calculationRequired( const QList<TreeItem*>& itemsToCalculateNow )
{
    if ( !top || top->getChildren().size() == 0 )
    {
        return false;
    }
    if ( itemsToCalculateNow.size() > 0 )
    {
        return true;
    }
    TreeItem* visibleTop = top->getChildren().first();
    if ( visibleTop->isCalculationEnabled() && !visibleTop->isCalculated() )
    {
        return true;
    }
    return false;
}

QVector<Task*>
SystemTree::setBasicValues( const QList<Tree*>&     leftTrees,
                            const QList<Tree*>&     rightTrees,
                            const QList<TreeItem*>& itemsToCalculate )
{
    QVector<Task*> workerData;
    /* No update of tree is necessary if the system tree is to the left of the metric tree */
    if ( ( leftTrees.size() == 0 ) || ( ( leftTrees.size() == 1 ) && ( leftTrees[ 0 ]->getTreeType() != METRICTREE ) ) )
    {
        return workerData;                   // the system tree is left of the metric tree.
    }

    if ( !calculationRequired( itemsToCalculate ) )
    {
        return workerData;
    }

    Tree* metricTree     = 0;
    Tree* callTree       = 0;
    Tree* callFlat       = 0;
    bool  rightOfCallDim = false;

    // Metric tree is definitely in the left trees, as the initial if-statement at the
    // beginning of the call already filtered out the other cases.
    foreach( Tree * item, leftTrees )
    {
        if ( item->getTreeType() == METRICTREE )
        {
            metricTree = item;
        }
        else if ( item->getTreeType() == CALLTREE )
        {
            callTree       = item;
            rightOfCallDim = true;
        }
        else if ( item->getTreeType() == CALLFLAT )
        {
            callFlat       = item;
            rightOfCallDim = true;
        }
    }
    assert( metricTree );

    // if call tree was not in the left trees, search in the right trees
    if ( !callTree && !callFlat )
    {
        foreach( Tree * item, rightTrees )
        {
            if ( item->getTreeType() == CALLTREE )
            {
                callTree = item;
            }
            else if ( item->getTreeType() == CALLFLAT )
            {
                callFlat = item;
            }
        }
    }
    assert( callTree || callFlat );

    // Set up selection for metric tree
    const QList<TreeItem*>& selected_metrics = metricTree->getSelectionList();
    list_of_metrics         metric_selection;

    foreach( TreeItem * item, selected_metrics )
    {
        metric_pair mp;
        mp.first  = static_cast<cube::Metric*> ( item->getCubeObject() );
        mp.second = ( item->isExpanded() ) ? CUBE_CALCULATE_EXCLUSIVE : CUBE_CALCULATE_INCLUSIVE;
        if ( item->getChildren().size() == 0 )
        {
            mp.second = CUBE_CALCULATE_INCLUSIVE;
        }
        metric_selection.push_back( mp );
    }

    // Set up selection for call tree
    if ( callTree )
    {
        list_of_cnodes cnode_selection;
        if ( rightOfCallDim )  // system tree is on the right
        {
            QList< TreeItem* > selected_cnodes = callTree->getSelectionList();
            foreach( TreeItem * item, selected_cnodes )
            {
                if ( item->isAggregatedLoopItem() )
                {
                    foreach( Cnode * contributingCnode, static_cast<AggregatedTreeItem*>( item )->getIterations() )
                    {
                        cnode_pair cp;
                        cp.first  = contributingCnode;
                        cp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                        cnode_selection.push_back( cp );
                    }
                }
                else
                {
                    cnode_pair cp;
                    cp.first  = static_cast<cube::Cnode*> ( item->getCubeObject() );
                    cp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                    cnode_selection.push_back( cp );
                }
            }
        }
        else
        {
            cnode_selection = CubeProxy::ALL_CNODES;
        }

        SystemTreeData* data = new SystemTreeData( this,
                                                   metric_selection,
                                                   cnode_selection );
        workerData.append( data );
    }


    // Set up selection for call flat
    else if ( callFlat )
    {
        // Set up selection for call flat
        list_of_regions region_selection;

        // Marks that some regions got status "subroutines" and have t be reset after calculation
        bool unset_regions = false;

        if ( rightOfCallDim )  // system tree is on the right
        {
            QList< TreeItem* > selected_regions = callFlat->getSelectionList();
            foreach( TreeItem * item, selected_regions )
            {
                region_pair rp;
                rp.first = static_cast<cube::Region*> ( item->getCubeObject() );
                if ( rp.first == NULL ) // subroutines
                {
                    unset_regions = true;
                    cube::Region* reg = static_cast<cube::Region*> ( item->getParent()->getCubeObject() );
                    reg->set_as_subroutines();
                    rp.first = reg;
                }
                rp.second = ( item->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;
                region_selection.push_back( rp );
            }
        }
        else
        {
            // Flat tree selection is the inclusive root node(s)
            QList< TreeItem* > selected_regions = callFlat->getTop().getChildren();
            TreeItem*          item             = selected_regions.at( 0 );
            region_pair        rp;
            rp.first  = static_cast<cube::Region*> ( item->getCubeObject() );
            rp.second = cube::CUBE_CALCULATE_INCLUSIVE;
            region_selection.push_back( rp );
        }


        SystemTreeFlatData* data = new SystemTreeFlatData( this,
                                                           metric_selection,
                                                           region_selection );
        workerData.append( data );

        if ( unset_regions )
        {
            for ( list_of_regions::iterator riter = region_selection.begin(); riter != region_selection.end(); ++riter )
            {
                ( *riter ).first->unset_as_subroutines();
            }
        }
    } // callflat

    computeMinMax();
    return workerData;
}
// end of computeBasicValues()

/**
 * adds the items of the given QList to list_of_cnodes. If the items are aggregated loop entries,
 * the nodes from whom the items are aggregated are added.
 */
void
SystemTree::fillCnodeList( QList<TreeItem*> items, list_of_cnodes& cv )
{
    foreach( TreeItem * treeItem, items )
    {
        cnode_pair cp;
        cp.second = ( treeItem->isExpanded() ) ? cube::CUBE_CALCULATE_EXCLUSIVE : cube::CUBE_CALCULATE_INCLUSIVE;

        if ( treeItem->isAggregatedRootItem() && treeItem->isExpanded() )
        {
            cp.first = static_cast<cube::Cnode*> ( treeItem->getCubeObject() );
            cv.push_back( cp );

            // add exclusive values of all hidden children of the root loop item
            cp.second = CUBE_CALCULATE_EXCLUSIVE;

            CallTree* callTree = static_cast<CallTree*>( treeItem->getTree() );
            foreach( TreeItem * iteration, callTree->getLoopRootItem()->getChildren() )
            {
                cp.first = static_cast<cube::Cnode*> ( iteration->cubeObject );
                cv.push_back( cp );
            }
        }
        else if ( !treeItem->isAggregatedLoopItem() ) // single item or collapsed root loop item
        {
            cp.first = static_cast<cube::Cnode*> ( treeItem->getCubeObject() );
            cv.push_back( cp );
        }
        else // aggregated loop item
        {
            AggregatedTreeItem* item = static_cast<AggregatedTreeItem*>( treeItem );
            foreach( Cnode * cnode, item->getIterations() )
            {
                if ( cnode ) // function has been called in current iteration
                {
                    cp.first = cnode;
                    cv.push_back( cp );
                }
            }
        }
    }
}

/**
 * @brief getVisitedItems
 * @return all threads with visited > 0
 */
QList<TreeItem*>
SystemTree::getVisitedItems()
{
    QList<TreeItem*> visited;

    Tree* metricTree = treeHash.value( METRICTREE );
    Tree* callTree   = treeHash.value( CALLTREE );
    if ( !callTree->isActive() )
    {
        callTree = treeHash.value( CALLFLAT );
    }

    TreeItem* visitedMetric = 0;
    foreach( TreeItem * item, metricTree->getItems() )
    {
        if ( item->getName() == "Visits" )
        {
            visitedMetric = item;
            break;
        }
    }

    // build up metric selection
    list_of_metrics metric_selection;
    metric_selection.push_back( make_pair( static_cast<Metric*>( visitedMetric->getCubeObject() ), CUBE_CALCULATE_INCLUSIVE ) );

    // build up cnode selection
    list_of_cnodes cnode_selection;
    foreach( TreeItem * callItem, callTree->getSelectionList() )
    {
        cube::CalculationFlavour callTreeState;
        if ( !callItem->isExpanded() )
        {
            callTreeState = cube::CUBE_CALCULATE_INCLUSIVE;
        }
        else
        {
            callTreeState = CUBE_CALCULATE_EXCLUSIVE;
        }
        cnode_selection.push_back( make_pair( static_cast<Cnode*>( callItem->getCubeObject() ), callTreeState ) );
    }

    vector<Value*> inclusive_values;
    vector<Value*> exclusive_values;

    // Get values for full tree. Inclusive values are needed as single threaded measurements
    // do not generate tree items for thread level and process level has 0 exclusive value.
    cube->getSystemTreeValues( metric_selection,
                               cnode_selection,
                               inclusive_values,
                               exclusive_values );

    foreach( TreeItem * item, getItems() )
    {
        if ( ( ( item->type == LOCATIONGROUPITEM && item->getChildren().size() == 0 )
               || ( item->type == LOCATIONITEM ) )
             && ( inclusive_values[ static_cast<Sysres*>( item->getCubeObject() )->get_sys_id() ]->getSignedLong() > 0 ) )
        {
            visited.append( item );
        }
    }

    return visited;
}


TreeItem*
SystemTree::getTreeItem( uint32_t systemId )
{
    foreach( TreeItem * item, treeItems )
    {
        cube::Sysres* sysres = static_cast<cube::Sysres*> ( item->getCubeObject() );
        if ( sysres->get_sys_id() == systemId )
        {
            return item;
        }
    }
    return 0;
}

void
SystemTreeData::calculate()
{
    // request bulk data from CubeProxy
    vector< Value* > inclusive_values;
    vector< Value* > exclusive_values;

    // get values for full tree
    cube::CubeProxy* cube = tree->getCube();
    cube->getSystemTreeValues( metric_selection,
                               cnode_selection,
                               inclusive_values,
                               exclusive_values );

    // Assign values to each toplevel tree
    foreach( TreeItem * child, tree->getTop().getChildren() )
    {
        tree->assignValues( child, inclusive_values, exclusive_values );
    }
    tree->computeMinMax();
}
void
SystemTreeFlatData::calculate()
{
    // request bulk data from CubeProxy
    vector< Value* > inclusive_values;
    vector< Value* > exclusive_values;

    // get values for full tree
    cube::CubeProxy* cube = tree->getCube();
    cube->getSystemTreeValues( metric_selection,
                               region_selection,
                               inclusive_values,
                               exclusive_values );

    // Assign values to each toplevel tree
    foreach( TreeItem * child, tree->getTop().getChildren() )
    {
        tree->assignValues( child, inclusive_values, exclusive_values );
    }
    tree->computeMinMax();
}
