/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2019                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>
#include <QList>
#include "ComputationPerformanceAnalysis.h"
#include "TreeItem.h"
#include "CubeAdvice.h"
#include "PerformanceTest.h"


using namespace cube;
using namespace cubegui;
using namespace cubepluginapi;
using namespace advisor;

ComputationPerformanceAnalysis::ComputationPerformanceAnalysis( cube::CubeProxy* _cube ) : PerformanceAnalysis( _cube )
{
    branch_test   = new       BranchPredictionPerformanceTest( cube );
    vector_test   = new       VectorizationTest( cube );
    l1_cache_test = new L1CacheUtilizationPerformanceTest( cube );
    l2_cache_test = new L2CacheUtilizationPerformanceTest( cube );
    l3_cache_test = new L3CacheUtilizationPerformanceTest( cube );
    cache_overall = new CachePerformanceTest( l1_cache_test,
                                              l2_cache_test,
                                              l3_cache_test );
    BP_ISSUE = tr( "One should improve branch prediction" );
    L1_ISSUE = tr( "Itilization of level 1 cache is not efficient" );
    L2_ISSUE = tr( "Itilization of level 2 cache is not efficient" );
    L3_ISSUE = tr( "Itilization of level-3 cache is not efficient" );
}


ComputationPerformanceAnalysis::~ComputationPerformanceAnalysis()
{
    delete branch_test;
    delete vector_test;
    delete l1_cache_test;
    delete l2_cache_test;
    delete l3_cache_test;
    delete cache_overall;
};



QList<AdvisorAdvice>
ComputationPerformanceAnalysis::getCandidates( const QList<TreeItem*>& _list )
{
    QList<AdvisorAdvice> to_return;
    foreach( TreeItem * item, _list )
    {
        cube::Cnode* cnode = static_cast<cube::Cnode*>( item->getCubeObject() );
        branch_test->apply( cnode, cube::CUBE_CALCULATE_EXCLUSIVE  );
        l1_cache_test->apply( cnode, cube::CUBE_CALCULATE_EXCLUSIVE  );
        l2_cache_test->apply( cnode, cube::CUBE_CALCULATE_EXCLUSIVE  );
        l3_cache_test->apply( cnode, cube::CUBE_CALCULATE_EXCLUSIVE  );

        double branch_value   = branch_test->value();
        double l1_cache_value = l1_cache_test->value();
        double l2_cache_value = l2_cache_test->value();
        double l3_cache_value = l3_cache_test->value();

        // if the values are exactly 0 they are most likely deactivated (e.g., MPI/OpenMP)
        // don't mark in these cases
        AdvisorAdvice advice;
        bool          mark          = false;
        QString       advice_string = tr( "Computation analysis: \n" );
        if ( branch_test->isActive() && ( branch_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += BP_ISSUE + QString( "\n" );
        }
        if ( l1_cache_test->isActive() && ( l1_cache_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += L1_ISSUE + QString( "\n" );
        }
        if ( l2_cache_test->isActive() && ( l2_cache_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += L2_ISSUE + QString( "\n" );
        }
        if ( l2_cache_test->isActive() &&  ( l3_cache_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += L3_ISSUE + QString( "\n" );
        }
        if ( mark )
        {
            advice.comments << advice_string;
            to_return << advice;
        }
    }
    return to_return;
}

QList<PerformanceTest*>
ComputationPerformanceAnalysis::getPerformanceTests()
{
    QList<PerformanceTest*> to_return;
    to_return << branch_test;
    to_return << vector_test;
    to_return <<  cache_overall;
    to_return << l1_cache_test;
    to_return << l2_cache_test;
    to_return << l3_cache_test;

    return to_return;
}


// ------ overview tests ---------

bool
ComputationPerformanceAnalysis::isActive() const
{
    return true;
}

QString
ComputationPerformanceAnalysis::getAnchorHowToMeasure()
{
    return "MeasurementForComputationAnalysis";
}
