/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2019                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>
#include <QList>
#include "OverviewPerformanceAnalysis.h"
#include "TreeItem.h"
#include "CubeAdvice.h"
#include "PerformanceTest.h"


using namespace cube;
using namespace cubegui;
using namespace cubepluginapi;
using namespace advisor;

OverviewPerformanceAnalysis::OverviewPerformanceAnalysis( cube::CubeProxy* _cube ) : PerformanceAnalysis( _cube )
{
    non_computation_load_balance = new NonComputationLoadBalanceTest( cube );
    computation_load_balance     = new ComputationLoadBalanceTest( cube );
    alg_mgm                      = new AlgMgmtTest( cube );
    communication                = new  CommunicationTest( cube );
    gpu_overhead                 = new  GPUOverheadTest( cube );
    io_overhead                  = new IOOverheadTest( cube );


    computation_overall =  new   OverallComputationPerformanceTest( computation_load_balance );
    management_overall  = new OverallManagementPerformanceTest( alg_mgm,
                                                                communication,
                                                                gpu_overhead,
                                                                io_overhead,
                                                                non_computation_load_balance     );

    overall             = new OverallPerformanceTest( computation_overall, management_overall );
    NON_COMP_LB_ISSUE   = tr( "Non computational code is not well balanced" );
    COMP_LB_ISSUE       = tr( "Computation is not well balanced" );
    ALG_MGM_ISSUE       = tr( "Algorithm management takes significant runtime" );
    COMMUNICATION_ISSUE = tr( "Communication takes significant runtime" );
}


OverviewPerformanceAnalysis::~OverviewPerformanceAnalysis()
{
    delete non_computation_load_balance;
    delete computation_load_balance;
    delete alg_mgm;
    delete communication;
    delete gpu_overhead;
    delete io_overhead;
    delete computation_overall;
    delete management_overall;
    delete overall;
};



QList<AdvisorAdvice>
OverviewPerformanceAnalysis::getCandidates( const QList<TreeItem*>& _list )
{
    QList<AdvisorAdvice> to_return;
    foreach( TreeItem * item, _list )
    {
        cube::Cnode* cnode = static_cast<cube::Cnode*>( item->getCubeObject() );
        non_computation_load_balance->apply( cnode, cube::CUBE_CALCULATE_INCLUSIVE  );
        computation_load_balance->apply( cnode, cube::CUBE_CALCULATE_INCLUSIVE  );
        alg_mgm->apply( cnode, cube::CUBE_CALCULATE_INCLUSIVE  );
        communication->apply( cnode, cube::CUBE_CALCULATE_INCLUSIVE  );

        double non_computation_load_balance_value = non_computation_load_balance->value();
        double computation_load_balance_value     = computation_load_balance->value();
        double alg_mgm_value                      = alg_mgm->value();
        double communication_value                = communication->value();

        // if the values are exactly 0 they are most likely deactivated (e.g., MPI/OpenMP)
        // don't mark in these cases
        AdvisorAdvice advice;
        bool          mark          = false;
        QString       advice_string = tr( "Overview analysis: \n" );
        if ( ( non_computation_load_balance_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += NON_COMP_LB_ISSUE + "\n";
        }
        if ( ( computation_load_balance_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += COMP_LB_ISSUE + "\n";
        }
        if ( ( alg_mgm_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += ALG_MGM_ISSUE + "\n";
        }
        if ( ( communication_value < 0.85 )  )
        {
            mark           |= true;
            advice.callpath = item;
            advice_string  += COMMUNICATION_ISSUE + "\n";
        }
        if ( mark )
        {
            advice.comments << advice_string;
            to_return << advice;
        }
    }
    return to_return;
}

QList<PerformanceTest*>
OverviewPerformanceAnalysis::getPerformanceTests()
{
    QList<PerformanceTest*> to_return;
    to_return << overall
              << management_overall
              << alg_mgm
              << communication
              << non_computation_load_balance
              << gpu_overhead
              << io_overhead
              << computation_overall
              << computation_load_balance;
    return to_return;
}


// ------ overview tests ---------

bool
OverviewPerformanceAnalysis::isActive() const
{
    return overall->isActive();
}

QString
OverviewPerformanceAnalysis::getAnchorHowToMeasure()
{
    return "MeasurementForOverviewAnalysis";
}
