/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2015-2020                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include <config.h>

#include "POPStalledResourcesTest.h"


using namespace advisor;

POPStalledResourcesTest::POPStalledResourcesTest( cube::CubeProxy* cube ) : PerformanceTest( cube )
{
    setName( tr( "Stalled Resources " ).toUtf8().data() );
    setWeight( 1 );   // need to be adjusted
    pop_stalled_resources = cube->getMetric( "stalled_resources" );
    if ( pop_stalled_resources == NULL )
    {
        adjustForTest( cube );
    }
    pop_stalled_resources = cube->getMetric( "stalled_resources" );
    if ( pop_stalled_resources == NULL )
    {
        setWeight( 0.1 );
        setValue( 0. );
        return;
    }

    cube::list_of_sysresources lsysres = getRootsOfSystemTree();
    cube::metric_pair          metric;
    metric.first  = pop_stalled_resources;
    metric.second = cube::CUBE_CALCULATE_INCLUSIVE;
    lmetrics.push_back( metric );
}



void
POPStalledResourcesTest::applyCnode( const cube::Cnode* cnode, const cube::CalculationFlavour cnf  )
{
    if ( pop_stalled_resources == NULL )
    {
        return;
    }

    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getCallpathSubtreeValues( lmetrics,
                                    lsysres,
                                    *( const_cast<cube::Cnode*>( cnode ) ),
                                    0,
                                    metric_id_indices,
                                    &inclusive_values,
                                    &exclusive_values );
    double pop_stalled_resources_value = ( cnf == cube::CUBE_CALCULATE_INCLUSIVE ) ?
                                         inclusive_values[ 0 ]->getDouble() :
                                         exclusive_values[ 0 ]->getDouble();
    setValue( pop_stalled_resources_value  );
}

void
POPStalledResourcesTest::applyCnode( const cube::list_of_cnodes& cnodes )
{
    if ( pop_stalled_resources == NULL )
    {
        return;
    }
    cube::value_container      inclusive_values;
    cube::value_container      exclusive_values;
    cube::list_of_sysresources lsysres;
    cube::IdIndexMap           metric_id_indices;
    cube->getMetricSubtreeValues( cnodes,
                                  lsysres,
                                  *pop_stalled_resources,
                                  0,
                                  metric_id_indices,
                                  &inclusive_values,
                                  &exclusive_values );
    double pop_stalled_resources_value = inclusive_values[ 0 ]->getDouble();
    setValue( pop_stalled_resources_value );
}



const std::string&
POPStalledResourcesTest::getCommentText() const
{
    if ( pop_stalled_resources == NULL )
    {
        return no_comment;
    }
    else
    {
        return no_comment;
    }
}


// ------ overview tests ---------

bool
POPStalledResourcesTest::isActive() const
{
    return pop_stalled_resources != NULL;
};

bool
POPStalledResourcesTest::isIssue() const
{
    return false;
};


void
POPStalledResourcesTest::adjustForTest( cube::CubeProxy* cube ) const
{
    cube::Metric* _res_stl_without_wait = cube->getMetric( "res_stl_without_wait" );
    cube::Metric* _tot_cyc_without_wait = cube->getMetric( "tot_cyc_without_wait" );

    if ( _res_stl_without_wait == NULL )
    {
        add_res_stl_without_wait( cube );
    }
    if ( _tot_cyc_without_wait == NULL )
    {
        add_tot_cyc_without_wait( cube );
    }
    _res_stl_without_wait = cube->getMetric( "res_stl_without_wait" );
    _tot_cyc_without_wait = cube->getMetric( "tot_cyc_without_wait" );
    if ( _res_stl_without_wait == NULL || _tot_cyc_without_wait == NULL )
    {
        return;
    }
    add_stalled_resources( cube );
}



void
POPStalledResourcesTest::add_stalled_resources( cube::CubeProxy* cube ) const
{
    cube::Metric* _met = cube->defineMetric(
        tr( "Stalled resources" ).toUtf8().data(),
        "stalled_resources",
        "DOUBLE",
        "",
        "",
        POP_STALLED_METRIC_URL,
        tr( "Amount of stalled resources, computed as res_stl_without_wait() / tot_cyc_without_wait()" ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_POSTDERIVED,
        " metric::res_stl_without_wait() / metric::tot_cyc_without_wait()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}



void
POPStalledResourcesTest::add_res_stl_without_wait( cube::CubeProxy* cube ) const
{
    cube::Metric* PAPI_TOT_INS = cube->getMetric( "PAPI_RES_STL" );
    if ( PAPI_TOT_INS == NULL )
    {
        return;
    }
    cube::Metric* _met = cube->defineMetric(
        tr( "PAPI_RES_STL without busy-wait, res_stl_without_wait" ).toUtf8().data(),
        "res_stl_without_wait",
        "DOUBLE",
        "",
        "",
        POP_IPC_METRIC_URL,
        tr( "Here is PAPI_RES_STL without busy-wait in MPI and OpenMP." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
        "${without_wait_state}[${calculation::callpath::id}] * metric::PAPI_RES_STL()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}



void
POPStalledResourcesTest::add_tot_cyc_without_wait( cube::CubeProxy* cube )  const
{
    cube::Metric* PAPI_TOT_CYC = cube->getMetric( "PAPI_TOT_CYC" );
    if ( PAPI_TOT_CYC == NULL )
    {
        return;
    }
    cube::Metric* _met = cube->defineMetric(
        tr( "PAPI_TOT_CYC without busy-wait" ).toUtf8().data(),
        "tot_cyc_without_wait",
        "DOUBLE",
        "",
        "",
        POP_IPC_METRIC_URL,
        tr( "Here is PAPI_TOT_CYC without busy-wait in MPI and OpenMP." ).toUtf8().data(),
        NULL,
        cube::CUBE_METRIC_PREDERIVED_EXCLUSIVE,
        "${without_wait_state}[${calculation::callpath::id}] * metric::PAPI_TOT_CYC()",
        "",
        "",
        "",
        "",
        true,
        cube::CUBE_METRIC_GHOST
        );
    if ( _met != NULL )
    {
        _met->setConvertible( false );
    }
    advisor_services->addMetric( _met, NULL );
}
